#ifndef __RDMAENDPOINT__
#define __RDMAENDPOINT__
#include <iostream>
#include <boost/lockfree/queue.hpp>
#include <rdma/rdma_cma.h>
#include <rdma/rdma_verbs.h>
#include <stdint.h>
#include <thread>
#include <netdb.h>
#include <arpa/inet.h>
#include <map>
#include <vector>
#include <mutex>
#include <shared_mutex>
#include "CqEventData.hpp"

class RdmaEndpoint
{
public:
    static int CONN_STATE_INITIALIZED;
    static int CONN_STATE_RESOURCES_ALLOCATED;
    static int CONN_STATE_CONNECTED;
    static int CONN_STATE_CLOSED;

    struct rdma_cm_id *_cm_id{NULL};
    struct ibv_cq *_completionQueue{NULL};
    struct ibv_pd *_protectionDomain{NULL};

    int _sendQueueSize{0};
    int _recvQueueSize{0};
    int _sendMsgSize{0};
    int _recvMsgSize{0};
    int _state{0};

    char *_sendBuff{NULL};
    char *_recvBuff{NULL};
    struct ibv_mr *_sendMr{NULL};
    struct ibv_mr *_recvMr{NULL};
    boost::lockfree::queue<void *> *_sendBuffers{NULL};

    RdmaEndpoint(struct rdma_cm_id *id, struct ibv_cq *completionQueue, int sendQueueSize,
                 int recvQueueSize, int sendMsgSize, int recvMsgSize);

    void createResources();
    void processCmEvent(struct rdma_cm_event *event);
    void clientClose();
    virtual void processSendCompletion(struct ibv_wc *data) = 0;
    virtual void processRecvCompletion(struct ibv_wc *data) = 0;
};
#endif