#include <rdma/rdma_cma.h>
#include <rdma/rdma_verbs.h>
#include <stdint.h>
#include <iostream>
#include <thread>
#include <unordered_map>

#ifndef __RDMAREPCQPROCESSOR__
#define __RDMAREPCQPROCESSOR__

#include "Executor.hpp"

class RdmaRepCqProcessor
{
public:
	struct ibv_comp_channel *_compChannel{NULL};
	struct ibv_cq *_completionQueue{NULL};
	std::thread *_compQueueThread{NULL};
	bool _stop{false};
	Executor *_executor{NULL};

	RdmaRepCqProcessor(Executor *ex, ibv_context *verbs, int compQueueSize)
		: _executor(ex)
	{
		_compChannel = ibv_create_comp_channel(verbs);
		if (_compChannel == NULL)
		{
			std::cout << "CqProcessr : ibv_create_comp_channel failed\n";
			return;
		}
		_completionQueue = ibv_create_cq(verbs, compQueueSize, NULL, _compChannel, 0);
		if (_completionQueue == NULL)
		{
			std::cout << "CqProcessr : ibv_create_cq failed" << std::endl;
			return;
		}
		int ret = ibv_req_notify_cq(_completionQueue, 0);
		if (ret)
		{
			std::cout << "CqProcessr : ibv_req_notify_cq failed\n";
		}
	}
	struct ibv_cq *getCq()
	{
		return _completionQueue;
	}
	void start()
	{
		std::cout << "CqProcessr : starting process CQ events" << std::endl;
		_compQueueThread = new std::thread(&RdmaRepCqProcessor::processCQEvents, this);
	}
	void processCQEvents()
	{
		int ret = 0;
		struct ibv_cq *cq;
		void *context;
		const int nevent = 10;
		struct ibv_wc *wc_array = new struct ibv_wc[nevent];
		while (!_stop)
		{
			ret = ibv_get_cq_event(_compChannel, &cq, &context);
			if (ret == -1)
			{
				std::cout << "CqProcessr : ibv_get_cq_event failed\n";
				close();
			}
			ibv_ack_cq_events(cq, 1);
			ret = ibv_req_notify_cq(_completionQueue, 0);
			if (ret)
			{
				std::cout << "CqProcessr : ibv_req_notify_cq failed\n";
				close();
			}
			ret = ibv_poll_cq(cq, nevent, wc_array);
			if (ret < 0)
			{
				std::cout << "CqProcessr : ibv_poll_cq failed\n";
				close();
			}
			if (ret == 0)
				continue;
			for (int i = 0; i < ret; i++)
			{
				struct ibv_wc *data = new struct ibv_wc(wc_array[i]);
				//data->vendor_err = 1;
				//_executor->submit(data);
				new std::thread(&RdmaRepCqProcessor::processRepEvent, this,data);
			}

			//_executor->dispatchRepCqEvents(wc_array, ret);
		}
	}
	void processRepEvent(struct ibv_wc* data)
	{
		std::cout<<"procesing Replication request"<<std::endl;
	}
	void close()
	{
		_stop = true;
		if (_compQueueThread != NULL)
			_compQueueThread->join();
	}
};

#endif
