#include "RdmaSalCqProcessor.hpp"
RdmaSalCqProcessor::RdmaSalCqProcessor(Executor *ex, ibv_context *verbs, int compQueueSize)
        : _executor(ex)
    {
        _compChannel = ibv_create_comp_channel(verbs);
        if (_compChannel == NULL)
        {
            std::cout << "SalCqProcessr : ibv_create_comp_channel failed\n";
            return;
        }
        _completionQueue = ibv_create_cq(verbs, compQueueSize, NULL, _compChannel, 0);
        if (_completionQueue == NULL)
        {
            std::cout << "SalCqProcessr : ibv_create_cq failed" << std::endl;
            return;
        }
        int ret = ibv_req_notify_cq(_completionQueue, 0);
        if (ret)
        {
            std::cout << "SalCqProcessr : ibv_req_notify_cq failed\n";
        }
    }
    struct ibv_cq *RdmaSalCqProcessor::getCq()
    {
        return _completionQueue;
    }
    void RdmaSalCqProcessor::start()
    {
        std::cout << "SalCqProcessr : starting process CQ events" << std::endl;
        _compQueueThread = new std::thread(&RdmaSalCqProcessor::processCQEvents, this);
    }
    void RdmaSalCqProcessor::processCQEvents()
    {
        int ret = 0;
        struct ibv_cq *cq;
        void *context;
        const int nevent = 10;
        struct ibv_wc wc_array[nevent];
        while (!_stop)
        {
            /*
            * get_CQ_event is a blocking call and it wait save some cpu cycles but.
            * it might not be that efficient compared to polling
            */
            ret = ibv_get_cq_event(_compChannel, &cq, &context);
            if (ret == -1)
            {
                std::cout << "SalCqProcessr : ibv_get_cq_event failed\n";
                close();
            }
            ibv_ack_cq_events(cq, 1);
            /* 
            * Create a request for next completion cycle
            */       
            ret = ibv_req_notify_cq(_completionQueue, 0);
            if (ret)
            {
                std::cout << "SalCqProcessr : ibv_req_notify_cq failed\n";
                close();
            }
            ret = ibv_poll_cq(cq, nevent, wc_array);
            if (ret < 0)
            {
                std::cout << "SalCqProcessr : ibv_poll_cq failed\n";
                close();
            }
            if (ret == 0)
                continue;
            for (int i = 0; i < ret; i++)
            {
                struct ibv_wc *data = new struct ibv_wc(wc_array[i]);
                /*
                * vendor_err is set to check whether the request came from sal or followers
                * data->vendor_err = 0;
                */
                _executor->submit(data);
            }
            //           _executor->dispatchSalCqEvents(wc_array, ret);
        }
    }

    void RdmaSalCqProcessor::close()
    {
        _stop = true;
        if (_compQueueThread != NULL)
            _compQueueThread->join();
    }