#include "RdmaSalEndpoint.hpp"
#include "MessageFormats.hpp"
RdmaSalEndpoint::RdmaSalEndpoint(struct rdma_cm_id *id, struct ibv_cq *completionQueue, int sendQueueSize,
                                 int recvQueueSize, int sendMsgSize, int recvMsgSize,rocksdb::DB *db)
    : RdmaEndpoint(id, completionQueue, sendQueueSize, recvQueueSize, sendMsgSize, recvMsgSize)
    ,_db(db)
{

}

void RdmaSalEndpoint::processSendCompletion(struct ibv_wc *data)
{
    std::cout << "send completion\n";
    _sendBuffers->push((void *)data->wr_id);
}

int RdmaSalEndpoint::sendMessage(const char *buffer, uint32_t size)
{
    if (size > _sendMsgSize)
        return -1;
    void* sendBuffer = nullptr;
    _sendBuffers->pop(sendBuffer);
    if (sendBuffer == nullptr)
        return -1;
    memcpy(sendBuffer, buffer, size);
    return rdma_post_send(_cm_id, sendBuffer, sendBuffer, size, _sendMr, 0);
}

void RdmaSalEndpoint::processRecvCompletion(struct ibv_wc *data)
{
    char *request = new char[data->byte_len];
    memcpy(request, (void *)data->wr_id, data->byte_len);
    struct SalRequest *req = (struct SalRequest *)request;
    rdma_post_recv(_cm_id, (void *)data->wr_id, (void *)data->wr_id, _recvMsgSize, _recvMr);

    std::cout << "recieve id : " << req->id << " " << (char *)req + SalRequestHeaderSize << " " << req->type << "size" << data->byte_len << "\n";

    if (req->type == RequestType::DELETE)
        processDelete(req);
    if (req->type == RequestType::GET)
        processGet(req);
    if (req->type == RequestType::PUT)
        processPut(req);
    delete[] request;
}

void RdmaSalEndpoint::processDelete(struct SalRequest *req)
{
    std::cout<<"0\n";
    rocksdb::Status s = _db->Delete(rocksdb::WriteOptions(), {(char *)req + SalRequestHeaderSize, req->keySize});
    std::cout<<"1\n";
    void *sendBuf = nullptr;
    _sendBuffers->pop(sendBuf);
    if (sendBuf == nullptr)
    {
        return;
    }
    char* sendBuffer = (char*)sendBuf;
    std::cout<<"2 "<<req->id<<"\n";
    memcpy(sendBuffer, &(req->id), sizeof(uint32_t));
    if (s.ok())
    {
        std::cout<<"33\n";
        memcpy(sendBuffer+4, &SUCCESS, sizeof(int));
        std::cout<<"44\n";
    }
    else
    {
        std::cout<<"331\n";
        memcpy(sendBuffer+4, &FAILURE, sizeof(int));
        std::cout<<"441\n";
    }
    rdma_post_send(_cm_id, sendBuffer, sendBuffer, _sendMsgSize, _sendMr, 0);
    std::cout<<"3\n";
}

void RdmaSalEndpoint::processGet(struct SalRequest *req)
{
    std::string value;
    rocksdb::Status s = _db->Get(rocksdb::ReadOptions(), {(char *)req + SalRequestHeaderSize, req->keySize}, &value);
    void *sendBuf = nullptr;
    _sendBuffers->pop(sendBuf);
    if (sendBuf == nullptr)
    {
        return;
    }

    char* sendBuffer = (char*)sendBuf;
    memcpy(sendBuffer, &(req->id), sizeof(uint32_t));
    if (s.ok())
    {
        memcpy(sendBuffer+4, &SUCCESS, sizeof(int));
        memcpy(sendBuffer+8, (void *)value.size(), sizeof(value.size()));
        memcpy(sendBuffer+12, value.c_str(), value.size());
    }
    else
    {
        memcpy(sendBuffer+4, &FAILURE, sizeof(int));
    }
    rdma_post_send(_cm_id, sendBuffer, sendBuffer, _sendMsgSize, _sendMr, 0);
}
void RdmaSalEndpoint::processPut(struct SalRequest *req)
{
    rocksdb::Status s = _db->Put(rocksdb::WriteOptions(), {(char *)req + SalRequestHeaderSize, req->keySize},
                                 {(char *)req + SalRequestHeaderSize + req->keySize, req->valueSize});
    void *sendBuf = nullptr;
    _sendBuffers->pop(sendBuf);
    if (sendBuf != nullptr)
    {
        return;
    }

    char* sendBuffer = (char*)sendBuf;
    memcpy(sendBuffer, &(req->id), sizeof(uint32_t));
    if (s.ok())
        memcpy(sendBuffer+4, &SUCCESS, sizeof(int));
    else
        memcpy(sendBuffer+4, &FAILURE, sizeof(int));
    rdma_post_send(_cm_id, sendBuffer, sendBuffer, _sendMsgSize, _sendMr, 0);
}

